"""
Test using the ``organize.organize`` function,
which is the primary interface for using the module.
"""
from organize import organize
from organize.tests import OrganizeTestCase


class TestOrganize(OrganizeTestCase):
    "Test organize function."

    def test_csv(self):
        "Test parsing a CSV."
        filename = 'csv/worldbank_preamble.csv'
        first_line = [
            (u'Country', u'Belarus'), (u'Year', u'2000'),
            (u'CO2 emissions (metric tons per capita)', u'5.91'),
            (u'Electric power consumption (kWh per capita)', u'2988.71'),
            (u'Energy use (kg of oil equivalent per capita)', u'2459.67'),
            (u'Fertility rate, total (births per woman)', u'1.29'),
            (u'GNI per capita, Atlas method (current US$)', u'1.38E+03'),
            (u'Internet users (per 1,000 people)', u'18.69'),
            (u'Life expectancy at birth, total (years)', u'68.01'),
            (u'Military expenditure (% of GDP)', u'1.26'),
            (u'Population, total', u'1.00E+07'),
            (u'Prevalence of HIV, total (% of population ages 15-49)', u''),
        ]

        with self.file_handle(filename) as fh:
            lines = organize(fh)
            first = list(lines.next())
            self.assertEquals(first, first_line)

    def test_tsv(self):
        "Test parsing a TSV."
        filename = 'tsv/imf_disb_repay.tsv'
        first_line = [
            (u'Year', u'2014'),
            (u'GRA Disbursements', u'2,765,465,000'),
            (u'GRA Repurchases', u'5,427,052,823'),
            (u'GRA Charges Paid', u'584,358,716'),
            (u'PRGT Disbursements', u'63,266,571'),
            (u'PRGT Repayments', u'132,553,847'),
            (u'PRGT Interest Paid', u'0'),
            (u'Total Disbursements', u'2,828,731,571'),
            (u'Total Repayments', u'5,559,606,670'),
            (u'Total Charges and Interest', u'584,358,716'),
        ]
        with self.file_handle(filename) as fh:
            lines = organize(fh)
            first = list(lines.next())
            self.assertEquals(first, first_line)

    def test_excel(self):
        "Test parsing an Excel."
        filename = 'excel/simple_whitespace.xlsx'
        first_line = [
            (u'Year', 1901),
            (u'Increase', 1),
            (u'Decrease', 14),
            (u'Grade', u'A'),
        ]
        with self.file_handle(filename) as fh:
            lines = organize(fh)
            first = list(lines.next())
            self.assertEquals(first, first_line)

    def test_to_dict(self):
        "Test transforming into dicts as described in README."
        filename = 'tsv/imf_disb_repay.tsv'
        first_dict = {
            u'PRGT Interest Paid': u'0',
            u'GRA Disbursements': u'2,765,465,000',
            u'GRA Charges Paid': u'584,358,716',
            u'Total Charges and Interest': u'584,358,716',
            u'PRGT Disbursements': u'63,266,571',
            u'Total Repayments': u'5,559,606,670',
            u'Year': u'2014',
            u'PRGT Repayments': u'132,553,847',
            u'GRA Repurchases': u'5,427,052,823',
            u'Total Disbursements': u'2,828,731,571',
        }

        with self.file_handle(filename) as fh:
            lines = organize(fh)
            first = dict(list(lines.next()))
            self.assertEquals(first, first_dict)



"Tab Separated Values parser."
import csv
import string
from organize.csv_parser import CSVParser
from itertools import islice


class TSVParser(CSVParser):
    """
    A parser which handles TSV data.

    A very lightweight subclass of ``CSVParser``.
    """
    mimetypes = ('text/tsv')
    extensions = ('.tsv')
    delimiter = '\t'
    default_dialect = csv.excel_tab

    def guess_dialect(self, txt):
        "CSV sniffer does not work well for TSV."
        return None

    def can_parse(self, stream):
        "Can parse this file as TSV."
        try:
            lines = islice(self.without_preamble(self.without_blank_lines(stream)), self.lines_to_test)
            for line in lines:
                if not string.count(line, self.delimiter):
                    return False
            return True
        except:
            return False



"Parser for Excel files."
"Comma Separated Values Parser."
from cStringIO import StringIO
from itertools import islice, chain, izip_longest
import xlrd
from organize.parsers import Parser


class ExcelParser(Parser):
    """
    A parser which handles Excel.

    Excel is pretty expressive, so we're basically just
    cherry-picking the first worksheet. This is clearly
    not a great long-term strategy.

    Right , so the sad truth is that xlrd does not play
    very well with data streams and wants to be able to
    reread the same data multiple times, so we will just
    have to oblige then, as painful as that happens to be
    """
    mimetypes = ('application/vnd.ms-excel', 'application/vs-excel')
    extensions = ('.xls', '.excel', '.xlsx', '.xlsm', '.xltm', '.xltx', '.xlsb')

    def file_contents(self, stream):
        """
        Return file contents for workbook.

        Ideally we'd do some kind of lazy stream reading
        which caches after it's read. This would allow us
        to avoid reading everything if we don't need it
        while also not requiring xlrd to know how to use
        streams properly.
        """
        return stream.read()

    def can_parse(self, stream):
        "Can parse this as an Excel file."
        try:
            xlrd.open_workbook(file_contents=self.file_contents(stream))
            return True
        except:
            return False

    def get_worksheet(self, workbook):
        """
        Pick the best worksheet from workbook.

        Later should consider using longest one rather
        than simple the first one.
        """
        for worksheet_name in workbook.sheet_names():
            return workbook.sheet_by_name(worksheet_name)

    def rows(self, sheet):
        "Filter out preamble from rows."

        num_cols_per_line = []
        start = 0
        previous = None
        for row_index in range(sheet.nrows):
            row = sheet.row_values(row_index)

            # this is probably a bit weak to simply use the
            # non-empty columns, perhaps a more robust check
            # would be to use the last empty column, although
            # in some situations that would likely not work either
            num_cols = len([x for x in row if x])
            if num_cols > 0:
                num_cols_per_line.append(num_cols)
                if previous is not None:
                    if num_cols > previous * 2:
                        start = row_index
                previous = num_cols

        headers = sheet.row_values(start)
        for row_index in range(start+1, sheet.nrows):
            cols = sheet.row_values(row_index)
            if len([x for x in cols if x]):
                yield izip_longest(headers, cols)

    def parse(self, stream):
        "Parse this file as Excel file."
        book = xlrd.open_workbook(file_contents=self.file_contents(stream))
        sheet = self.get_worksheet(book)
        return self.rows(sheet)



"""
Parse a variety of tabular data formats.

    >>> from organize import organize
    >>> for row in organize(open('myfile', 'r')):
            for col, val in row:
                print "%s: %s" % (col, val)

"""
from organize import parsers


class Organizer(object):
    """
    ``Organizer`` determines which formats to attempt to
    parse files with, and the order in which to parse them in.

    """
    def __init__(self, filename=None, mimetype=None, use_default_parsers=True, prepend_parsers=None, append_parsers=None, use_filename_hints=True, use_mimetype_hints=True):
        """
        Initialize Organizer.

        ``use_default_parsers`` is True or False, and if False does not load any standard parsers (CSV, etc),
            and you will only have the parsers you've listed in ``prepend_parsers`` and ``append_parsers``.
        ``prepend_parsers`` is a list of parser classes to try before default parsers
        ``append_parsers`` is a list of parser classes to try after default parsers
        ``use_filename_hints`` re-sorts the parsers to first try parsers that can use the specified filename or extension
        ``use_mimetype_hints`` re-sorts the parsers to first try parsers that can use the specified mimetype
        """
        prepend = prepend_parsers if prepend_parsers else []
        default = self.default_parsers() if use_default_parsers else []
        append = append_parsers if append_parsers else []
        parsers = prepend + default + append
        self.parsers = [parser() for parser in parsers]

    def default_parsers(self):
        "Find parsers for Organizer to support current formats."
        return parsers.registered_parsers()

    def determine_parser_order(self, filename=None, mimetype=None):
        "Determine order to attempt parsers."
        parser_order = []
        if filename:
            for parser in self.parsers:
                if parser.should_parse_by_filename(filename) and parser not in parser_order:
                    parser_order.append(parser)
        if mimetype:
            for parser in self.parsers:
                if parser.should_parse_by_mimetype(mimetype) and parser not in parser_order:
                    parser_order.append(parser)

        for parser in self.parsers:
            if parser not in parser_order:
                parser_order.append(parser)

        return parser_order

    def parse(self, stream, filename=None, mimetype=None):
        """
        Parse a file stream.

        ``stream`` is a file-like resource containing the data
        """
        parser_order = self.determine_parser_order(filename=filename, mimetype=mimetype)
        exception = None
        for parser in parser_order:
            stream.seek(0)
            try:
                if parser.can_parse(stream):
                    stream.seek(0)
                    return parser.parse(stream)
            except Exception, e:
                exception = e
        if exception:
            raise exception  # pylint: disable-msg=E0702


def organize(stream, filename=None, mimetype=None):
    """
    Read a file stream.

        >>> with open('mydata.csv', 'r') as fin:
                organize(fin)

    """
    return Organizer().parse(stream, filename=filename, mimetype=mimetype)



"""
Holds base Parser class and the parser registry of default parsers.
"""
import os.path
from organize.utils import import_class


# Generally it's better to use ``append_parsers`` or ``prepend_parsers``
# in ``Optimizer.__init__`` than to modify the default parser registry.
PARSER_REGISTRY = ['organize.csv_parser.CSVParser', 'organize.tsv_parser.TSVParser', 'organize.excel_parser.ExcelParser']


def registered_parsers():
    "Retrieve parser classes."
    return [import_class(x) for x in PARSER_REGISTRY]


class Parser(object):
    "Base parser class."
    mimetypes = ()
    extensions = ()
    delimiter = ''

    def can_parse(self, stream):
        """
        Determine if you can read a stream,
        while reading as little as possible.

        Returns True or False

        Does not need to reset seek.
        """
        raise NotImplementedError

    def parse(self, stream):
        """
        Create a generator which transforms the stream into rows.

        Specifically it should yield each row as a list of
        2-tuples (column name, column value).
        """
        raise NotImplementedError

    def should_parse_by_filename(self, filename):
        """
        Return whether or not the parser believes it can
        parse a file based on the file's filename.

        This is used to prioritize parsers, and will
        not prevent attempting a parser on a poorly
        labeled file.
        """
        ext = os.path.splitext(filename)[1] if filename else None
        return ext in self.extensions

    def should_parse_by_mimetype(self, mimetype):
        """
        Whether or not the parser believes it can parse
        a file based on the file's mimetype.

        This is used to prioritize parsers, and will
        not prevent attempting a parser on a poorly
        labeled file.
        """
        return mimetype in self.mimetypes

    def without_blank_lines(self, stream):
        "Filter out blank lines."
        to_strip = '\t\r\n %s' % self.delimiter
        for line in stream:
            if line.strip(to_strip):
                yield line



"Init file for tests directory."
import unittest
import os.path


class OrganizeTestCase(unittest.TestCase):
    """
    ``unittest.TestCase`` subclass with a few extra
    utility methods to simplify testing.
    """
    data_path = os.path.join(os.path.dirname(__file__), 'data')
    csv_filenames = ['csv/Baby_Names__Beginning_2007.csv',  'csv/worldbank.csv']
    tsv_filenames = ['tsv/imf_disb_repay.tsv']
    excel_filenames = ['excel/irrat_exu_robert_shiller.xlsx', 'excel/simple.xlsx', 'excel/simple_old.xls']

    def file_handle(self, filename):
        "Create file handle for a data file."
        path = os.path.join(self.data_path, filename)
        return open(path, 'r')



"Comma Separated Values Parser."
import csv
from itertools import islice, chain, izip_longest
from organize.parsers import Parser
from cStringIO import StringIO


def unicode_csv_reader(unicode_csv_data, dialect=csv.excel, **kwargs):
    """
    csv.py doesn't do Unicode; encode temporarily as UTF-8

    See: https://docs.python.org/2.7/library/csv.html
    """
    csv_reader = csv.reader(utf_8_encoder(unicode_csv_data), dialect=dialect, **kwargs)
    for row in csv_reader:
        yield [unicode(cell, 'utf-8') for cell in row]


def utf_8_encoder(unicode_csv_data):
    """
    Work around for csv module's utf-8 issues.

    See: https://docs.python.org/2.7/library/csv.html
    """
    for line in unicode_csv_data:
        yield line.encode('utf-8')


class CSVParser(Parser):
    """
    A parser which handles CSV. This is a bit different than
    the Python CSV module which is really a *SV module in disguise.

    As such, this CSVParser is intentionally more restrictive to
    try to address situations where a file is valid CSV and valid
    TSV at the same time (but one of them is coherent and the other
    is largely incoherent).

    ``organize.tsv.TSVParser`` lightly subclasses this class.
    """
    mimetypes = ('text/csv')
    extensions = ('.csv')
    delimiter = ','
    default_dialect = csv.excel
    lines_to_test = 25

    def guess_dialect(self, txt):
        "Try to guess CSV dialect."
        utf8_txt = txt.encode('utf-8')
        try:
            return csv.Sniffer().sniff(utf8_txt, delimiters=self.delimiter)
        except:
            pass

    def can_parse(self, stream):
        "Can parse this file as CSV."
        try:
            lines = islice(self.without_preamble(self.without_blank_lines(stream)), self.lines_to_test)
            txt = '\n'.join(lines)
            dialect = self.guess_dialect(txt)
            if dialect and self.delimiter in txt:
                return True
            return False
        except:
            return False

    def without_preamble(self, stream):
        """
        Remove a stream's preamble if any.

        This is necessarily pretty hacky. What we do here is parse
        the first few lines for columns until we stabilize on a given
        number of rows. If early rows have a different number of columns
        then we cut them off.
        """
        lines = list(islice(stream, self.lines_to_test))
        num_cols_per_line = []
        start = 0
        previous = None
        for i, line in enumerate(lines):
            num_cols = len(list(csv.reader(StringIO(line), delimiter=self.delimiter))[0])
            num_cols_per_line.append(num_cols)
            if previous is not None:
                if num_cols > previous * 2:
                    start = i
            previous = num_cols

        for line in chain(lines[start:], stream):
            yield line

    def parse(self, stream):
        """
        Parse this file as CSV.
        """
        lines = self.without_preamble(self.without_blank_lines(stream))
        lines_for_dialect = list(islice(lines, self.lines_to_test))
        dialect = self.guess_dialect('\n'.join(lines_for_dialect))
        if not dialect:
            dialect = self.default_dialect

        # we need to chain the lines we read for dialect detection together
        # with the rest of the unread lines to avoid rereading or dropping
        # their contents
        reader = unicode_csv_reader(chain(lines_for_dialect, lines), dialect=dialect)

        # use first row for headers
        headers = list(islice(reader, 0, 1))[0]

        for row in reader:
            yield ((name, val) for name, val in izip_longest(headers, row))



"Test TSV parsing."
from itertools import islice
import organize.tsv_parser
from organize.tests import OrganizeTestCase


class TestTSVParser(OrganizeTestCase):
    "Test TSV."
    def setUp(self):
        "Test TSV parser."
        self.parser = organize.tsv_parser.TSVParser()
        self.first_line = [
            (u'Year', u'2014'),
            (u'GRA Disbursements', u'2,765,465,000'),
            (u'GRA Repurchases', u'5,427,052,823'),
            (u'GRA Charges Paid', u'584,358,716'),
            (u'PRGT Disbursements', u'63,266,571'),
            (u'PRGT Repayments', u'132,553,847'),
            (u'PRGT Interest Paid', u'0'),
            (u'Total Disbursements', u'2,828,731,571'),
            (u'Total Repayments', u'5,559,606,670'),
            (u'Total Charges and Interest', u'584,358,716'),
        ]

    def test_should_parse_by_mimetype(self):
        "Responds to proper mimetype."
        self.assertTrue(self.parser.should_parse_by_mimetype('text/tsv'))
        self.assertFalse(self.parser.should_parse_by_mimetype('text/csv'))
        self.assertFalse(self.parser.should_parse_by_mimetype('application/vs-excel'))

    def test_should_parse_by_filename(self):
        "Responds to proper extensions."
        self.assertTrue(self.parser.should_parse_by_filename('myfile/test.tsv'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.csv'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.xls'))

    def test_can_parse(self):
        "Test determining which files are parsable."
        for filename in self.tsv_filenames:
            with self.file_handle(filename) as fh:
                self.assertTrue(self.parser.can_parse(fh))

        for filename in self.csv_filenames + self.excel_filenames:
            with self.file_handle(filename) as fh:
                self.assertFalse(self.parser.can_parse(fh))

    def test_parse(self):
        "Test parsing TSV files."
        filename = 'tsv/imf_disb_repay.tsv'
        with self.file_handle(filename) as fh:
            first = list(self.parser.parse(fh).next())
            self.assertEquals(first, self.first_line)



"Some simple utilities."
import importlib


def import_class(class_path):
    "Import a class using full path."
    parts = class_path.split('.')
    assert len(parts) > 1, "class path must include a module and a class, for example organize.csv.CSVParser"
    module_path = ".".join(parts[:-1])
    class_name = parts[-1]
    module_obj = importlib.import_module(module_path)
    return getattr(module_obj, class_name)



from distutils.core import setup
import setuptools

setup(
    name='organize',
    version='0.1.2',
    author='Will Larson',
    author_email='lethain@gmail.com',
    packages=['organize', 'organize.tests'],
    url='http://pypi.python.org/pypi/organize/',
    license='LICENSE.txt',
    description='Parse real-world tabular data in a wide variety of formats.',
    long_description=open('README.md').read(),
    install_requires=['xlrd'],
)



"Test CSV parsing."
from itertools import islice
import organize.csv_parser
from organize.tests import OrganizeTestCase


class TestCSVParser(OrganizeTestCase):
    "Test CSV."
    def setUp(self):
        "Test CSV parser."
        self.parser = organize.csv_parser.CSVParser()
        self.worldbank_first_line = [
            (u'Country', u'Belarus'), (u'Year', u'2000'),
            (u'CO2 emissions (metric tons per capita)', u'5.91'),
            (u'Electric power consumption (kWh per capita)', u'2988.71'),
            (u'Energy use (kg of oil equivalent per capita)', u'2459.67'),
            (u'Fertility rate, total (births per woman)', u'1.29'),
            (u'GNI per capita, Atlas method (current US$)', u'1.38E+03'),
            (u'Internet users (per 1,000 people)', u'18.69'),
            (u'Life expectancy at birth, total (years)', u'68.01'),
            (u'Military expenditure (% of GDP)', u'1.26'),
            (u'Population, total', u'1.00E+07'),
            (u'Prevalence of HIV, total (% of population ages 15-49)', u''),
        ]

    def test_should_parse_by_mimetype(self):
        "Responds to proper mimetype."
        self.assertTrue(self.parser.should_parse_by_mimetype('text/csv'))
        self.assertFalse(self.parser.should_parse_by_mimetype('application/vs-excel'))
        self.assertFalse(self.parser.should_parse_by_mimetype('text/tsv'))

    def test_should_parse_by_filename(self):
        "Responds to proper extensions."
        self.assertTrue(self.parser.should_parse_by_filename('test.csv'))
        self.assertTrue(self.parser.should_parse_by_filename('myfile/test.csv'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.xls'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.tsv'))

    def test_can_parse(self):
        "Test determining which files are parsable."
        for filename in self.csv_filenames:
            with self.file_handle(filename) as fh:
                self.assertTrue(self.parser.can_parse(fh))

        for filename in self.tsv_filenames + self.excel_filenames:
            with self.file_handle(filename) as fh:
                self.assertFalse(self.parser.can_parse(fh))

    def test_parse(self):
        "Test parsing CSV files."
        filename = 'csv/worldbank.csv'
        with self.file_handle(filename) as fh:
            first = list(self.parser.parse(fh).next())
            self.assertEquals(first, self.worldbank_first_line)

    def test_parse_despite_empty_lines(self):
        "Test parsing CSV files with empty lines scattered about."
        filename = 'csv/worldbank_whitespace.csv'
        with self.file_handle(filename) as fh:
            lines = self.parser.parse(fh)
            first = list(lines.next())
            self.assertEquals(first, self.worldbank_first_line)
            self.assertEquals(3, len(list(lines)))

    def test_parse_despite_preamble(self):
        "Test parsing CSV files with empty lines scattered about and a preamble."
        filename = 'csv/worldbank_preamble.csv'
        with self.file_handle(filename) as fh:
            lines = self.parser.parse(fh)
            first = list(lines.next())
            self.assertEquals(first, self.worldbank_first_line)
            self.assertEquals(3, len(list(lines)))



"Test Excel parsing."
import organize.excel_parser
from itertools import islice
from organize.tests import OrganizeTestCase


class TestExcelParser(OrganizeTestCase):
    "Test Excel."
    def setUp(self):
        "Test Excel parser."
        self.parser = organize.excel_parser.ExcelParser()
        self.first_line = [
            (u'Year', 1901),
            (u'Increase', 1),
            (u'Decrease', 14),
            (u'Grade', u'A'),
        ]

    def test_should_parse_by_mimetype(self):
        "Responds to proper mimetype."
        self.assertTrue(self.parser.should_parse_by_mimetype('application/vs-excel'))
        self.assertTrue(self.parser.should_parse_by_mimetype('application/vnd.ms-excel'))
        self.assertFalse(self.parser.should_parse_by_mimetype('text/csv'))
        self.assertFalse(self.parser.should_parse_by_mimetype('text/tsv'))

    def test_should_parse_by_filename(self):
        "Responds to proper extensions."
        self.assertTrue(self.parser.should_parse_by_filename('myfile/test.xls'))
        self.assertTrue(self.parser.should_parse_by_filename('myfile/test.xlsx'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.csv'))
        self.assertFalse(self.parser.should_parse_by_filename('myfile/test.tsv'))

    def test_can_parse(self):
        "Test determining which files are parsable."
        for filename in self.excel_filenames:
            with self.file_handle(filename) as fh:
                self.assertTrue(self.parser.can_parse(fh))

        for filename in self.tsv_filenames + self.csv_filenames:
            with self.file_handle(filename) as fh:
                self.assertFalse(self.parser.can_parse(fh))

    def test_parse_xlsx(self):
        "Test parsing Excel files."
        filename = 'excel/simple.xlsx'
        with self.file_handle(filename) as fh:
            first = list(self.parser.parse(fh).next())
            self.assertEquals(first, self.first_line)

    def test_parse_xls(self):
        "Test parsing Excel-97 files."
        filename = 'excel/simple_old.xls'
        with self.file_handle(filename) as fh:
            first = list(self.parser.parse(fh).next())
            self.assertEquals(first, self.first_line)

    def test_parse_skip_whitespace(self):
        "Test parsing Excel files."
        filename = 'excel/simple_whitespace.xlsx'
        with self.file_handle(filename) as fh:
            first = list(self.parser.parse(fh).next())
            self.assertEquals(first, self.first_line)

    def test_parse_with_preamble(self):
        "Test parsing a file with a preamble."
        filename = "excel/irrat_exu_robert_shiller.xlsx"
        num_rows = 0
        with self.file_handle(filename) as fh:
            for row in self.parser.parse(fh):
                row_lst = list(row)
                self.assertTrue(len(row_lst) >= 10, row_lst)
                num_rows += 1
        self.assertTrue(num_rows > 1500)
